@@ -1,6 +1,6 @@
 /* -*- Mode: C; c-file-style: "gnu" -*-
  * udev-configure-printer - a udev callout to configure print queues
- * Copyright (C) 2009, 2010 Red Hat, Inc.
+ * Copyright (C) 2009, 2010, 2011 Red Hat, Inc.
  * Author: Tim Waugh <twaugh@redhat.com>
  *
  * This program is free software; you can redistribute it and/or modify
@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
  *
  */
 
@@ -45,13 +45,89 @@
 #include <ctype.h>
 #include <syslog.h>
 #include <unistd.h>
-#include <usb.h>
+#include <libusb.h>
 #include <glib.h>
+#include <dirent.h>
 
 #define DISABLED_REASON "Unplugged or turned off"
 #define MATCH_ONLY_DISABLED 1
 #define USB_URI_MAP "/var/run/udev-configure-printer/usb-uris"
 
+#if (CUPS_VERSION_MAJOR > 1) || (CUPS_VERSION_MINOR > 5)
+#define HAVE_CUPS_1_6 1
+#endif
+
+/*
+ * CUPS 1.6 makes various structures private and
+ * introduces these ippGet and ippSet functions
+ * for all of the fields in these structures.
+ * http://www.cups.org/str.php?L3928
+ * We define (same signatures) our own accessors when CUPS < 1.6.
+ */
+#ifndef HAVE_CUPS_1_6
+const char *
+ippGetName(ipp_attribute_t *attr)
+{
+  return (attr->name);
+}
+
+ipp_op_t
+ippGetOperation(ipp_t *ipp)
+{
+  return (ipp->request.op.operation_id);
+}
+
+ipp_status_t
+ippGetStatusCode(ipp_t *ipp)
+{
+  return (ipp->request.status.status_code);
+}
+
+ipp_tag_t
+ippGetGroupTag(ipp_attribute_t *attr)
+{
+  return (attr->group_tag);
+}
+
+ipp_tag_t
+ippGetValueTag(ipp_attribute_t *attr)
+{
+  return (attr->value_tag);
+}
+
+int
+ippGetInteger(ipp_attribute_t *attr,
+              int             element)
+{
+  return (attr->values[element].integer);
+}
+
+const char *
+ippGetString(ipp_attribute_t *attr,
+             int             element,
+             const char      **language)
+{
+  return (attr->values[element].string.text);
+}
+
+ipp_attribute_t	*
+ippFirstAttribute(ipp_t *ipp)
+{
+  if (!ipp)
+    return (NULL);
+  return (ipp->current = ipp->attrs);
+}
+
+ipp_attribute_t *
+ippNextAttribute(ipp_t *ipp)
+{
+  if (!ipp || !ipp->current)
+    return (NULL);
+  return (ipp->current = ipp->current->next);
+}
+
+#endif
+
 struct device_uris
 {
   size_t n_uris;
@@ -441,6 +517,12 @@ parse_device_id (const char *device_id,
   free (fieldname);
 }
 
+int
+device_file_filter(const struct dirent *entry)
+{
+  return ((strstr(entry->d_name, "lp") != NULL) ? 1 : 0);
+}
+
 static char *
 device_id_from_devpath (const char *devpath,
 			const struct usb_uri_map *map,
@@ -451,20 +533,27 @@ device_id_from_devpath (const char *devpath,
   struct usb_uri_map_entry *entry;
   struct udev *udev;
   struct udev_device *dev, *parent_dev = NULL;
-  const char *sys;
   const char *idVendorStr, *idProductStr, *serial;
   char *end;
   unsigned long idVendor, idProduct;
   size_t syslen, devpathlen;
-  char *syspath;
-  struct usb_bus *bus;
-  struct usb_dev_handle *handle = NULL;
+  char *syspath, *devicefilepath;
+  libusb_device **list;
+  struct libusb_device	*device;
+  struct libusb_device_handle *handle = NULL;
+  struct libusb_device_descriptor devdesc;
+  struct libusb_config_descriptor *confptr = NULL;
+  const struct libusb_interface *ifaceptr = NULL;
+  const struct libusb_interface_descriptor *altptr = NULL;
+  char libusbserial[1024];
   char ieee1284_id[1024];
   const char *device_id = NULL;
-  int conf = 0, iface = 0;
+  int conf = 0, iface = 0, altset = 0, numdevs = 0, i, n, m;
   int got = 0;
   char *usb_device_devpath;
   char *usblpdevpos, *dest;
+  struct dirent **namelist;
+  int num_names;
 
   id->full_device_id = id->mfg = id->mdl = id->sern = NULL;
 
@@ -475,33 +564,7 @@ device_id_from_devpath (const char *devpath,
       exit (1);
     }
 
-  /* For devices discovered via the usblp kernel module we read out the number
-   * of the /dev/usb/lp* device file, as there can be queues set up with 
-   * non-standard CUPS backends based on the /dev/usb/lp* device file and
-   * we want to avoid that an additional queue with a standard CUPS backend
-   * gets set up.
-   */
-  usblpdevpos = strstr(devpath, "/usb/lp");
-  if (usblpdevpos != NULL)
-    usblpdevpos += 7;
-  else
-    {
-      usblpdevpos = strstr(devpath, "/usblp");
-      if (usblpdevpos != NULL)
-	usblpdevpos += 6;
-    }
-  if (usblpdevpos != NULL)
-    {
-      for (dest = usblpdev;
-	   (*usblpdevpos >= '0') && (*usblpdevpos <= '9') &&
-	     (dest - usblpdev < usblpdevlen);
-	   usblpdevpos ++, dest ++)
-	*dest = *usblpdevpos;
-      *dest = '\0';
-    }
-
-  sys = udev_get_sys_path (udev);
-  syslen = strlen (sys);
+  syslen = strlen ("/sys");
   devpathlen = strlen (devpath);
   syspath = malloc (syslen + devpathlen + 1);
   if (syspath == NULL)
@@ -510,18 +573,52 @@ device_id_from_devpath (const char *devpath,
       syslog (LOG_ERR, "out of memory");
       exit (1);
     }
-
-  memcpy (syspath, sys, syslen);
+  memcpy (syspath, "/sys", syslen);
   memcpy (syspath + syslen, devpath, devpathlen);
   syspath[syslen + devpathlen] = '\0';
 
+  devicefilepath = malloc (syslen + devpathlen + 5);
+  if (devicefilepath == NULL)
+    {
+      udev_unref (udev);
+      syslog (LOG_ERR, "out of memory");
+      exit (1);
+    }
+  memcpy (devicefilepath, syspath, syslen + devpathlen);
+  memcpy (devicefilepath + syslen + devpathlen, "/usb", 4);
+  devicefilepath[syslen + devpathlen + 4] = '\0';
+
+  /* For devices under control of the usblp kernel module we read out the number
+   * of the /dev/usb/lp* device file, as there can be queues set up with 
+   * non-standard CUPS backends based on the /dev/usb/lp* device file and
+   * we want to avoid that an additional queue with a standard CUPS backend
+   * gets set up.
+   */
+  num_names = scandir(devicefilepath, &namelist, device_file_filter, alphasort);
+  if (num_names <= 0)
+    num_names = scandir(syspath, &namelist, device_file_filter, alphasort);
+  if (num_names > 0)
+    {
+      usblpdevpos = strstr(namelist[0]->d_name, "lp");
+      if (usblpdevpos != NULL)
+	{
+	  usblpdevpos += 2;
+	  for (dest = usblpdev;
+	       (*usblpdevpos >= '0') && (*usblpdevpos <= '9') &&
+		 (dest - usblpdev < usblpdevlen);
+	       usblpdevpos ++, dest ++)
+	    *dest = *usblpdevpos;
+	  *dest = '\0';
+	}
+    }
+
   dev = udev_device_new_from_syspath (udev, syspath);
   if (dev == NULL)
     {
       udev_device_unref (dev);
       udev_unref (udev);
       syslog (LOG_ERR, "unable to access %s", syspath);
-      exit (1);
+      return NULL;
     }
 
   parent_dev = udev_device_get_parent_with_subsystem_devtype (dev,
@@ -531,28 +628,31 @@ device_id_from_devpath (const char *devpath,
     {
       udev_unref (udev);
       syslog (LOG_ERR, "Failed to get parent");
-      exit (1);
+      return NULL;
     }
 
   usb_device_devpath = strdup (udev_device_get_devpath (parent_dev));
-  syslog (LOG_DEBUG, "parent devpath is %s", usb_device_devpath);
+  syslog (LOG_DEBUG, "device devpath is %s", usb_device_devpath);
 
   for (entry = map->entries; entry; entry = entry->next)
     if (!strcmp (entry->devpath, usb_device_devpath))
       break;
 
   if (entry)
-    /* The map already had an entry so has already been dealt
-     * with.  This can happen because there are two "add"
-     * triggers: one for the usb_device device and the other for
-     * the usblp device.  We have most likely been triggered by
-     * the usblp device, so the usb_device rule got there before
-     * us and succeeded.
-     *
-     * Pretend we didn't find any device URIs that matched, and
-     * exit.
-     */
-    exit (0);
+    {
+      /* The map already had an entry so has already been dealt
+       * with.  This can happen because there are two "add"
+       * triggers: one for the usb_device device and the other for
+       * the usblp device.  We have most likely been triggered by
+       * the usblp device, so the usb_device rule got there before
+       * us and succeeded.
+       *
+       * Pretend we didn't find any device URIs that matched, and
+       * exit.
+       */
+      syslog (LOG_DEBUG, "Device already handled");
+      return NULL;
+    }
 
   serial = udev_device_get_sysattr_value (parent_dev, "serial");
   if (serial)
@@ -563,15 +663,17 @@ device_id_from_devpath (const char *devpath,
   else
     usbserial[0] = '\0';
 
-  /* See if we were triggered by a usblp add event. */
-  device_id = udev_device_get_sysattr_value (dev, "device/ieee1284_id");
+  /* See if  the device is controlled by the usblp kernel module
+     and so the module has already read the device ID for us */
+  device_id = udev_device_get_sysattr_value (dev, "ieee1284_id");
   if (device_id)
     {
       got = 1;
       goto got_deviceid;
     }
 
-  /* This is a low-level USB device.  Use libusb to fetch the Device ID. */
+  /* The usblp kernel module is not attached to this device. Use libusb to
+     fetch the Device ID. */
   idVendorStr = udev_device_get_sysattr_value (parent_dev, "idVendor");
   idProductStr = udev_device_get_sysattr_value (parent_dev, "idProduct");
 
@@ -582,119 +684,127 @@ device_id_from_devpath (const char *devpath,
       syslog (LOG_ERR, "Missing sysattr %s",
 	      idVendorStr ?
 	      (idProductStr ? "serial" : "idProduct") : "idVendor");
-      exit (1);
+      return NULL;
     }
 
   idVendor = strtoul (idVendorStr, &end, 16);
   if (end == idVendorStr)
     {
       syslog (LOG_ERR, "Invalid idVendor: %s", idVendorStr);
-      exit (1);
+      return NULL;
     }
 
   idProduct = strtoul (idProductStr, &end, 16);
   if (end == idProductStr)
     {
       syslog (LOG_ERR, "Invalid idProduct: %s", idProductStr);
-      exit (1);
+      return NULL;
     }
 
   syslog (LOG_DEBUG, "Device vendor/product is %04zX:%04zX",
 	  idVendor, idProduct);
 
-  usb_init ();
-  usb_find_busses ();
-  usb_find_devices ();
-  for (bus = usb_get_busses (); bus && !got; bus = bus->next)
-    {
-      struct usb_device *device;
-      for (device = bus->devices; device && !got; device = device->next)
-	{
-	  struct usb_config_descriptor *confptr;
-	  if (device->descriptor.idVendor != idVendor ||
-	      device->descriptor.idProduct != idProduct ||
-	      !device->config)
-	    continue;
+  libusb_init(NULL);
+  numdevs = libusb_get_device_list(NULL, &list);
+  if (numdevs > 0)
+    for (i = 0; i < numdevs; i++)
+      {
+	device = list[i];
 
-	  conf = 0;
-	  for (confptr = device->config;
-	       conf < device->descriptor.bNumConfigurations && !got;
-	       conf++, confptr++)
-	    {
-	      struct usb_interface *ifaceptr;
-	      iface = 0;
-	      for (ifaceptr = confptr->interface;
-		   iface < confptr->bNumInterfaces && !got;
-		   iface++, ifaceptr++)
-		{
-		  struct usb_interface_descriptor *altptr;
-		  int altset = 0;
-		  for (altptr = ifaceptr->altsetting;
-		       altset < ifaceptr->num_altsetting && !got;
-		       altset++, altptr++)
-		    {
-		      if (altptr->bInterfaceClass == USB_CLASS_PRINTER &&
-			  altptr->bInterfaceSubClass == 1)
-			{
-			  int n;
-			  handle = usb_open (device);
-			  if (!handle)
-			    {
-			      syslog (LOG_DEBUG, "failed to open device");
-			      continue;
-			    }
+	if (libusb_get_device_descriptor(device, &devdesc) < 0)
+	  continue;
 
-			  n = altptr->bInterfaceNumber;
-			  if (usb_claim_interface (handle, n) < 0)
-			    {
-			      usb_close (handle);
-			      handle = NULL;
-			      syslog (LOG_DEBUG, "failed to claim interface");
-			      continue;
-			    }
+	if (!devdesc.bNumConfigurations || !devdesc.idVendor ||
+	    !devdesc.idProduct)
+	  continue;
 
-			  if (n != 0 && usb_claim_interface (handle, 0) < 0)
-			    {
-			      usb_close (handle);
-			      handle = NULL;
-			      syslog (LOG_DEBUG, "failed to claim interface 0");
-			      continue;
-			    }
+	if (devdesc.idVendor != idVendor || devdesc.idProduct != idProduct)
+	  continue;
 
-			  n = altptr->bAlternateSetting;
-			  if (usb_set_altinterface (handle, n) < 0)
-			    {
-			      usb_close (handle);
-			      handle = NULL;
-			      syslog (LOG_DEBUG, "failed set altinterface");
-			      continue;
-			    }
+	for (conf = 0; conf < devdesc.bNumConfigurations; conf ++)
+	  {
+	    if (libusb_get_config_descriptor(device, conf, &confptr) < 0)
+	      continue;
+	    for (iface = 0, ifaceptr = confptr->interface;
+		 iface < confptr->bNumInterfaces;
+		 iface ++, ifaceptr ++)
+	      {
+		for (altset = 0, altptr = ifaceptr->altsetting;
+		     altset < ifaceptr->num_altsetting;
+		     altset ++, altptr ++)
+		  {
+		    if (altptr->bInterfaceClass != LIBUSB_CLASS_PRINTER ||
+			altptr->bInterfaceSubClass != 1)
+		      continue;
+		    
+		    if (libusb_open(device, &handle) < 0)
+		      {
+			syslog (LOG_DEBUG, "failed to open device");
+			continue;
+		      }
 
-			  memset (ieee1284_id, '\0', sizeof (ieee1284_id));
-			  if (usb_control_msg (handle,
-					       USB_TYPE_CLASS |
-					       USB_ENDPOINT_IN |
-					       USB_RECIP_INTERFACE,
-					       0, conf, iface,
-					       ieee1284_id,
-					       sizeof (ieee1284_id),
-					       5000) < 0)
-			    {
-			      usb_close (handle);
-			      handle = NULL;
-			      syslog (LOG_ERR, "Failed to fetch Device ID");
-			      continue;
-			    }
+		    if (usbserial[0] != '\0' &&
+			(libusb_get_string_descriptor_ascii(handle,
+						devdesc.iSerialNumber,
+						(unsigned char *)libusbserial,
+						sizeof(libusbserial))) > 0 &&
+			strcmp(usbserial, libusbserial) != 0)
+		      {
+			libusb_close (handle);
+			handle = NULL;
+			continue;
+		      }
+		      
+		    n = altptr->bInterfaceNumber;
+		    if (libusb_claim_interface(handle, n) < 0)
+		      {
+			libusb_close (handle);
+			handle = NULL;
+			syslog (LOG_DEBUG, "failed to claim interface");
+			continue;
+		      }
+		    if (n != 0 && libusb_claim_interface(handle, 0) < 0)
+		      {
+			syslog (LOG_DEBUG, "failed to claim interface 0");
+		      }
+		    
+		    m = altptr->bAlternateSetting;
+		    if (libusb_set_interface_alt_setting(handle, n, m)
+			< 0)
+		      {
+			libusb_close (handle);
+			handle = NULL;
+			syslog (LOG_DEBUG, "failed set altinterface");
+			continue;
+		      }
 
-			  got = 1;
-			  usb_close (handle);
-			  break;
-			}
-		    }
-		}
-	    }
-	}
-    }
+		    memset (ieee1284_id, '\0', sizeof (ieee1284_id));
+		    if (libusb_control_transfer(handle,
+						LIBUSB_REQUEST_TYPE_CLASS |
+						LIBUSB_ENDPOINT_IN |
+						LIBUSB_RECIPIENT_INTERFACE,
+						0, conf,
+						(n << 8) | m,
+						(unsigned char *)ieee1284_id,
+						sizeof (ieee1284_id),
+						5000) < 0)
+		      {
+			libusb_close (handle);
+			handle = NULL;
+			syslog (LOG_ERR, "Failed to fetch Device ID");
+			continue;
+		      }
+
+		    got = 1;
+		    libusb_close (handle);
+		    break;
+		  }
+	      }
+	  }
+      }
+
+  libusb_free_device_list(list, 1);
+  libusb_exit(NULL);
 
  got_deviceid:
   if (got)
@@ -766,15 +876,15 @@ cupsDoRequestOrDie (http_t *http,
   if (answer == NULL)
     {
       syslog (LOG_ERR, "failed to send IPP request %d",
-	      request->request.op.operation_id);
+	      ippGetOperation (request));
       exit (1);
     }
 
-  if (answer->request.status.status_code > IPP_OK_CONFLICT)
+  if (ippGetStatusCode (answer) > IPP_OK_CONFLICT)
     {
       syslog (LOG_ERR, "IPP request %d failed (%d)",
-	      request->request.op.operation_id,
-	      answer->request.status.status_code);
+	      ippGetOperation (request),
+	      ippGetStatusCode (answer));
       exit (1);
     }
 
@@ -820,15 +930,12 @@ find_matching_device_uris (struct device_id *id,
   cups = httpConnectEncrypt (cupsServer (), ippPort(), cupsEncryption ());
   if (cups == NULL)
     {
-      /* Don't bother retrying here.  Instead, the CUPS initscript
-	 should run these commands after cupsd is started:
-
-	 udevadm trigger --subsystem-match=usb \
-	                 --attr-match=bInterfaceClass=07 \
-			 --attr-match=bInterfaceSubClass=01
-
-	 udevadm trigger --subsystem-match=usb \
-	                 --property-match=DEVNAME="/dev/usb/lp*"
+      /* Don't bother retrying here.  We've probably been run from
+	 udev before the cups.socket systemd unit is running.  We'll
+	 get run again, as the systemd service
+	 udev-configure-printer.service, after cups.socket.  For more
+	 information:
+	 http://0pointer.de/blog/projects/socket-activation2.html
       */
 
       syslog (LOG_DEBUG, "failed to connect to CUPS server; giving up");
@@ -845,26 +952,26 @@ find_matching_device_uris (struct device_id *id,
   answer = cupsDoRequestOrDie (cups, request, "/");
   httpClose (cups);
 
-  for (attr = answer->attrs; attr; attr = attr->next)
+  for (attr = ippFirstAttribute (answer); attr; attr = ippNextAttribute (answer))
     {
       const char *device_uri = NULL;
       struct device_id this_id;
       this_id.full_device_id = this_id.mfg = this_id.mdl = this_id.sern = NULL;
 
-      while (attr && attr->group_tag != IPP_TAG_PRINTER)
-	attr = attr->next;
+      while (attr && ippGetGroupTag (attr) != IPP_TAG_PRINTER)
+	attr = ippNextAttribute (answer);
 
       if (!attr)
 	break;
 
-      for (; attr && attr->group_tag == IPP_TAG_PRINTER; attr = attr->next)
+      for (; attr && ippGetGroupTag (attr) == IPP_TAG_PRINTER; attr = ippNextAttribute (answer))
 	{
-	  if (attr->value_tag == IPP_TAG_URI &&
-	      !strcmp (attr->name, "device-uri"))
-	    device_uri = attr->values[0].string.text;
-	  else if (attr->value_tag == IPP_TAG_TEXT &&
-		   !strcmp (attr->name, "device-id"))
-	    parse_device_id (attr->values[0].string.text, &this_id);
+	  if (ippGetValueTag (attr) == IPP_TAG_URI &&
+	      !strcmp (ippGetName (attr), "device-uri"))
+	    device_uri = ippGetString (attr, 0, NULL);
+	  else if (ippGetValueTag (attr) == IPP_TAG_TEXT &&
+		   !strcmp (ippGetName (attr), "device-id"))
+	    parse_device_id (ippGetString (attr, 0, NULL), &this_id);
 	}
 
       /* Only use device schemes in our preference order for matching
@@ -1158,6 +1265,7 @@ for_each_matching_queue (struct device_uris *device_uris,
     "printer-state-message",
   };
   char usblpdevstr1[32] = "", usblpdevstr2[32] = "";
+  int firstqueue = 1;
 
   if (cups == NULL)
     return 0;
@@ -1175,9 +1283,9 @@ for_each_matching_queue (struct device_uris *device_uris,
       exit (1);
     }
 
-  if (answer->request.status.status_code > IPP_OK_CONFLICT)
+  if (ippGetStatusCode (answer) > IPP_OK_CONFLICT)
     {
-      if (answer->request.status.status_code == IPP_NOT_FOUND)
+      if (ippGetStatusCode (answer) == IPP_NOT_FOUND)
 	{
 	  /* No printer queues configured. */
 	  ippDelete (answer);
@@ -1185,7 +1293,7 @@ for_each_matching_queue (struct device_uris *device_uris,
 	}
 
       syslog (LOG_ERR, "CUPS-Get-Printers request failed (%d)",
-	      answer->request.status.status_code);
+	      ippGetStatusCode (answer));
       exit (1);
     }
 
@@ -1200,12 +1308,9 @@ for_each_matching_queue (struct device_uris *device_uris,
 	       usblpdev);
       snprintf(usblpdevstr2, sizeof(usblpdevstr2), "/usblp%s",
 	       usblpdev);
-      syslog (LOG_DEBUG,
-	      "Consider also queues with \"%s\" or \"%s\" in their URIs as matching",
-	      usblpdevstr1, usblpdevstr2);
     }
 
-  for (attr = answer->attrs; attr; attr = attr->next)
+  for (attr = ippFirstAttribute (answer); attr; attr = ippNextAttribute (answer))
     {
       const char *this_printer_uri = NULL;
       const char *this_device_uri = NULL;
@@ -1215,32 +1320,35 @@ for_each_matching_queue (struct device_uris *device_uris,
       char *this_device_uri_n, *device_uri_n;
       const char *ps1, *ps2, *pi1, *pi2;
 
-      while (attr && attr->group_tag != IPP_TAG_PRINTER)
-	attr = attr->next;
+      while (attr && ippGetGroupTag (attr) != IPP_TAG_PRINTER)
+	attr = ippNextAttribute (answer);
 
       if (!attr)
 	break;
 
-      for (; attr && attr->group_tag == IPP_TAG_PRINTER; attr = attr->next)
+      for (; attr && ippGetGroupTag (attr) == IPP_TAG_PRINTER; attr = ippNextAttribute (answer))
 	{
-	  if (attr->value_tag == IPP_TAG_URI)
+	  if (ippGetValueTag (attr) == IPP_TAG_URI)
 	    {
-	      if (!strcmp (attr->name, "device-uri"))
-		this_device_uri = attr->values[0].string.text;
-	      else if (!strcmp (attr->name, "printer-uri-supported"))
-		this_printer_uri = attr->values[0].string.text;
+	      if (!strcmp (ippGetName (attr), "device-uri"))
+		this_device_uri = ippGetString (attr, 0, NULL);
+	      else if (!strcmp (ippGetName (attr), "printer-uri-supported"))
+		this_printer_uri = ippGetString (attr, 0, NULL);
 	    }
-	  else if (attr->value_tag == IPP_TAG_TEXT &&
-		   !strcmp (attr->name, "printer-state-message"))
-	    printer_state_message = attr->values[0].string.text;
-	  else if (attr->value_tag == IPP_TAG_ENUM &&
-		   !strcmp (attr->name, "printer-state"))
-	    state = attr->values[0].integer;
+	  else if (ippGetValueTag (attr) == IPP_TAG_TEXT &&
+		   !strcmp (ippGetName (attr), "printer-state-message"))
+	    printer_state_message = ippGetString (attr, 0, NULL);
+	  else if (ippGetValueTag (attr) == IPP_TAG_ENUM &&
+		   !strcmp (ippGetName (attr), "printer-state"))
+	    state = ippGetInteger (attr, 0);
 	}
 
+      if (!this_device_uri)
+	/* CUPS didn't include a device-uri attribute in the response
+	   for this printer (shouldn't happen). */
+	goto skip;
+
       this_device_uri_n = normalize_device_uri(this_device_uri);
-      syslog (LOG_DEBUG, "URI of print queue: %s, normalized: %s",
-	      this_device_uri, this_device_uri_n);
       pi1 = strstr (this_device_uri, "interface=");
       ps1 = strstr (this_device_uri, "serial=");
       for (i = 0; i < device_uris->n_uris; i++)
@@ -1251,10 +1359,10 @@ for_each_matching_queue (struct device_uris *device_uris,
 	     level USB (libusb) we cannot simply compare URIs, must
 	     consider also URIs as equal if one has an "interface"
 	     or "serial" attribute and the other not. If both have
-	     the attribute it must naturally match. We check which attributes 
-             are there and this way determine up to which length the two URIs 
+	     the attribute it must naturally match. We check which attributes
+             are there and this way determine up to which length the two URIs
              must match. Here we can assume that if a URI has an "interface"
-	  `  attribute it has also a "serial" attribute, as this URI is
+	     attribute it has also a "serial" attribute, as this URI is
 	     an URI obtained via libusb and these always have a "serial"
 	     attribute. usblp-based URIs never have an "interface"
 	     attribute.*/
@@ -1272,16 +1380,26 @@ for_each_matching_queue (struct device_uris *device_uris,
 	    l = strlen(this_device_uri);
 	  else
 	    l = strlen(device_uris->uri[i]);
-	  syslog (LOG_DEBUG, "URI of detected printer: %s, normalized: %s",
-		  device_uris->uri[i], device_uri_n);
+	  if (firstqueue == 1)
+	    {
+	      syslog (LOG_DEBUG, "URI of detected printer: %s, normalized: %s",
+		      device_uris->uri[i], device_uri_n);
+	      if (i == 0 && strlen(usblpdev) > 0)
+		syslog (LOG_DEBUG,
+			"Consider also queues with \"%s\" or \"%s\" in their URIs as matching",
+			usblpdevstr1, usblpdevstr2);
+	    }
+	  if (i == 0)
+	    syslog (LOG_DEBUG, "URI of print queue: %s, normalized: %s",
+		    this_device_uri, this_device_uri_n);
 	  if ((!strncmp (device_uris->uri[i], this_device_uri, l)) ||
-	      (strstr (device_uri_n, this_device_uri_n) == 
+	      (strstr (device_uri_n, this_device_uri_n) ==
 	       device_uri_n) ||
-	      (strstr (this_device_uri_n, device_uri_n) == 
+	      (strstr (this_device_uri_n, device_uri_n) ==
 	       this_device_uri_n) ||
 	      ((strlen(usblpdev) > 0) &&
 	       ((strstr (this_device_uri, usblpdevstr1) != NULL) ||
-	       (strstr (this_device_uri, usblpdevstr2) != NULL)))) 
+	       (strstr (this_device_uri, usblpdevstr2) != NULL))))
 	    {
 	      matched++;
 	      syslog (LOG_DEBUG, "Queue %s has matching device URI",
@@ -1292,10 +1410,14 @@ for_each_matching_queue (struct device_uris *device_uris,
 		  (flags & MATCH_ONLY_DISABLED) == 0)
 		{
 		  (*fn) (this_printer_uri, context);
+		  break;
 		}
 	    }
 	}
 
+      firstqueue = 0;
+
+    skip:
       if (!attr)
 	break;
     }
@@ -1326,7 +1448,7 @@ enable_queue (const char *printer_uri, void *context)
       return;
     }
 
-  if (answer->request.status.status_code > IPP_OK_CONFLICT)
+  if (ippGetStatusCode (answer) > IPP_OK_CONFLICT)
     syslog (LOG_ERR, "IPP-Resume-Printer request failed");
   else
     syslog (LOG_INFO, "Re-enabled printer %s", printer_uri);
@@ -1368,7 +1490,7 @@ bluetooth_verify_address (const char *bdaddr)
 }
 
 static int
-do_add (const char *cmd, const char *devpath)
+do_add (const char *cmd, const char *devpath, int detach)
 {
   pid_t pid;
   int f;
@@ -1380,7 +1502,7 @@ do_add (const char *cmd, const char *devpath)
   char usblpdev[8] = "";
   gboolean is_bluetooth;
 
-  if (getenv ("DEBUG") == NULL)
+  if (detach && getenv ("DEBUG") == NULL)
     {
       if ((pid = fork ()) == -1)
 	syslog (LOG_ERR, "Failed to fork process");
@@ -1417,12 +1539,7 @@ do_add (const char *cmd, const char *devpath)
   }
 
   if (!id.mfg || !id.mdl)
-    {
-      syslog (LOG_DEBUG, "invalid or missing IEEE 1284 Device ID%s%s",
-	      id.full_device_id ? " " : "",
-	      id.full_device_id ? id.full_device_id : "");
-      exit (1);
-    }
+    return 1;
 
   syslog (LOG_DEBUG, "MFG:%s MDL:%s SERN:%s serial:%s", id.mfg, id.mdl,
 	  id.sern ? id.sern : "-", usbserial[0] ? usbserial : "-");
@@ -1518,7 +1635,7 @@ remove_queue (const char *printer_uri)
       return;
     }
 
-  if (answer->request.status.status_code > IPP_OK_CONFLICT)
+  if (ippGetStatusCode (answer) > IPP_OK_CONFLICT)
     syslog (LOG_ERR, "IPP-Delete-Printer request failed");
   else
     syslog (LOG_INFO, "Deleted printer %s as the corresponding device "
@@ -1552,7 +1669,7 @@ disable_queue (const char *printer_uri, void *context)
       return;
     }
 
-  if (answer->request.status.status_code > IPP_OK_CONFLICT)
+  if (ippGetStatusCode (answer) > IPP_OK_CONFLICT)
     syslog (LOG_ERR, "IPP-Pause-Printer request failed");
   else
     syslog (LOG_INFO, "Disabled printer %s as the corresponding device "
@@ -1607,26 +1724,103 @@ do_remove (const char *devpath)
   return 0;
 }
 
+static int
+do_enumerate (const char *argv0)
+{
+  struct udev *udev;
+  struct udev_enumerate *e;
+  struct udev_list_entry *item = NULL, *first = NULL;
+  int r = 0;
+
+  udev = udev_new ();
+  if (udev == NULL)
+    {
+      syslog (LOG_ERR, "udev_new failed");
+      exit (1);
+    }
+
+  e = udev_enumerate_new (udev);
+  if (e == NULL)
+    {
+      udev_unref (udev);
+      syslog (LOG_ERR, "udev_enumerate_new failed");
+      exit (1);
+    }
+
+  if (udev_enumerate_add_match_tag (e, "udev-configure-printer") < 0)
+    {
+      udev_unref (udev);
+      udev_enumerate_unref (e);
+      syslog (LOG_ERR, "udev_enumerate_add_match_tag failed");
+      exit (1);
+    }
+
+  if (udev_enumerate_scan_devices (e) < 0)
+    {
+      udev_unref (udev);
+      udev_enumerate_unref (e);
+      syslog (LOG_ERR, "udev_enumerate_scan_devices failed");
+      exit (1);
+    }
+
+  first = udev_enumerate_get_list_entry (e);
+  udev_list_entry_foreach (item, first)
+    {
+      int k;
+      const char *p;
+      struct udev_device *d;
+
+      d = udev_device_new_from_syspath (udev, udev_list_entry_get_name(item));
+      if (!d)
+        continue;
+
+      p = udev_device_get_devpath (d);
+      if (p)
+        {
+          k = do_add (argv0, p, 0);
+          if (k != 0)
+            r = k;
+        }
+
+      udev_device_unref (d);
+  }
+
+  udev_enumerate_unref (e);
+  udev_unref (udev);
+
+  return r;
+}
+
 int
 main (int argc, char **argv)
 {
-  int add;
+  int add = 0;
+  int enumerate = 0;
 
-  if (argc != 3 ||
-      !((add = !strcmp (argv[1], "add")) ||
-	!strcmp (argv[1], "remove")))
+  if (argc > 1)
+    {
+      add = !strcmp (argv[1], "add");
+      enumerate = !strcmp (argv[1], "enumerate");
+    }
+
+  if (!(argc == 3 &&
+        (add || !strcmp (argv[1], "remove"))) &&
+      !(argc == 2 && enumerate))
     {
       fprintf (stderr,
 	       "Syntax: %s add {USB device path}\n"
-	       "        %s remove {USB device path}\n",
-	       argv[0], argv[0]);
+	       "        %s remove {USB device path}\n"
+               "        %s enumerate\n",
+	       argv[0], argv[0], argv[0]);
       return 1;
     }
 
   openlog ("udev-configure-printer", 0, LOG_LPR);
   cupsSetPasswordCB (no_password);
   if (add)
-    return do_add (argv[0], argv[2]);
+    return do_add (argv[0], argv[2], 1);
+  if (enumerate)
+    return do_enumerate (argv[0]);
 
   return do_remove (argv[2]);
 }
